/**
 * This is a sequential program that tests 335544320000 graphs to see if they 
 * can help us find a tighter bound on R(5,5).
 */

#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

using namespace std;

// Constants 
const long long NUM_GRAPHS = 335544320000LL; // A long long has a maximum value of 9223372036854775807.
const int RED = 0;
const int BLUE = 1;
const int NUM_VERTICES = 45;
const int NUM_EDGES = 990; 

// Function Prototypes
void generateNextGraph(int graph[]);
bool hasK5(int graph[]);

// Main
int main(int argc, char *argv[])
{
	// Start timing how long it takes the program to check all the graphs.
	time_t start = time(NULL);//starts timer

	int graph[NUM_EDGES]; // The graph in each set of graphs to test.
		
	// Initializing the arrays
	for(int j=0; j < NUM_EDGES; j++)
	{
		graph[j]=0;
	}//end for

	// Tests the first graph.
	bool result = hasK5(graph);

	if(result == false)
	{
		for (int l = 0; l < NUM_EDGES; l++) 
		{
			cout << graph[l];
		}
		cout<<endl;
	}//end if


	long long loopIndex = 0LL;

	 // This is a for loop that tests all graphs for this core.
	for(loopIndex = 0; loopIndex < NUM_GRAPHS; loopIndex++)
	{
		// Generate next graph.
		generateNextGraph(graph);

		// Test graph.
		result = hasK5(graph);

		// If a graph with no K5 is found, print out that graph.
		if(result == false)
		{
			for (int l = 0; l < NUM_EDGES; l++) 
			{
				cout << graph[l];
			}
			cout<<endl;
		}//end if
	}//end for loop


	time_t stop = time(NULL);//ends timer
	int delta = stop - start;//calculates the difference of start and stop time
	cout << "Time taken " << delta <<" seconds " << endl;//outputs time (in seconds)

	return 0;
}//end main


/* This function tests the graph passed in to see if it contains a K5.
 * @param graph - The array holding the graph.
 * @return - Returns whether the graph contains a K5.
 */
bool hasK5(int graph[])
{
	bool foundK5 = false;

	// initialize counters
	int vertex1 = 0;
	int vertex2 = 1;
	int vertex3 = 2;
	int vertex4 = 3;
	int vertex5 = 4;

	/* These five nested while loops select the 5 vertices that 
	   we will test to see if they form a K5. The outermost while loop
	   selects the smallest numbered vertex.The next while loop selets the second smallest
	   numbered vertex and so on... The set of while loops test every possible
	   set of five vertices that could form a K5 until it finds one or has tested every possible
	   set.
	*/

	while (!foundK5 && vertex1 < NUM_VERTICES - 4)
	{
		while (!foundK5 && vertex2 < NUM_VERTICES - 3)
		{
			while (!foundK5 && vertex3 < NUM_VERTICES - 2)
			{
				while (!foundK5 && vertex4 < NUM_VERTICES - 1)
				{
					while (!foundK5 && vertex5 < NUM_VERTICES)
					{
						if (graph[((NUM_VERTICES * vertex1) + vertex2 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
							graph[((NUM_VERTICES * vertex1) + vertex3 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
							graph[((NUM_VERTICES * vertex1) + vertex4 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
							graph[((NUM_VERTICES * vertex1) + vertex5 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&							
							graph[((NUM_VERTICES * vertex2) + vertex3 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
							graph[((NUM_VERTICES * vertex2) + vertex4 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
							graph[((NUM_VERTICES * vertex2) + vertex5 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
							graph[((NUM_VERTICES * vertex3) + vertex4 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == RED &&							
							graph[((NUM_VERTICES * vertex3) + vertex5 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == RED &&
							graph[((NUM_VERTICES * vertex4) + vertex5 - (((vertex4 + 1)*(vertex4 + 2))/ 2))] == RED)
						{
							foundK5 = true;
						}
						else if (graph[((NUM_VERTICES * vertex1) + vertex2 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE&&
							graph[((NUM_VERTICES * vertex1) + vertex3 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex1) + vertex4 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex1) + vertex5 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex2) + vertex3 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex2) + vertex4 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex2) + vertex5 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex3) + vertex4 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex3) + vertex5 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == BLUE &&
							graph[((NUM_VERTICES * vertex4) + vertex5 - (((vertex4 + 1)*(vertex4 + 2))/ 2))] == BLUE)
						{
							foundK5 = true;
						}
						vertex5++;
					}
					vertex4++;
					vertex5 = vertex4 + 1;
				}
				vertex3++;
				vertex4 = vertex3 + 1;
				vertex5 = vertex4 + 1;
			}
			vertex2++;
			vertex3 = vertex2 + 1;
			vertex4 = vertex3 + 1;
			vertex5 = vertex4 + 1;
		}
		vertex1++;
		vertex2 = vertex1 + 1;
		vertex3 = vertex2 + 1;
		vertex4 = vertex3 + 1;
		vertex5 = vertex4 + 1;
	}

	if (foundK5 == true)
	{
		return true;
	}
	else
	{
		return false;
	}
}


/* This function generates the next graph.
 * @param graph - The array holding the graph.
 */
void generateNextGraph( int graph[])
{
	bool carry = false;
	int index = NUM_EDGES -1;	

	do 
	{
   		//add 1 and check for carry 
  		if ( graph[index] == 0)
   		{
			carry = false;
			graph[index] = 1; //set to 1
		}
  		else
		{
      		carry = true;
			graph[index] = 0; //set to 0
		}//end if/else

		index = index - 1;
	}
    while((carry == true) && (index >= 0));

	index = NUM_EDGES - 1;
}

