#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <time.h>
#include <windows.h>

using namespace std;

#include "template_kernel.cu"
#include <cutil_inline.h>
#include <template_kernel.cu>

extern "C"

int main(int argc, char **argv)
{
	// ********START TIMING********
	time_t start = time(NULL);//starts timer

	int N = 990;
	
	size_t size = N * sizeof(int);

	//*************ALLOCATE MEMORY FOR HOST**********

    // starting graph for GPU to use
	int* mySourceGraph_host = (int*) malloc(size); 

    // place for solution if we find it on host
    int* mySolutionGraph_host = (int*) malloc(size); 

    // flag to tell us if we find a graph
    bool foundSolution_host;

    // This array holds 0 in each slot when initialized.  It will hold the number+1 of
    // the graph that contains no red or blue K5 (if found) after the GPU executes the
    // kernel and the array is copied from the GPU.  In other words, if a slot in the 
    // array holds the value x which is greater than 0, then subtract 1 from x to get
    // the graph (aka 0 to 19,999 or 0 to 9,999) number.  Add that to the (array index *
    // 10,000 or 20,000 depending on how many graphs each thread tests).  Then convert that
    // to binary and voila! you have the graph that shows R(5,5) >= 46.
    int* answer_host = (int*) malloc(16777216 * sizeof(int));

    // Fill the array with zeros.  This function prevents an optimizing compiler from
    // optimizing out this step which can occur with ZeroMemory().
    SecureZeroMemory(answer_host, sizeof(answer_host));

    //***********************************************
	
	cout << "memory allocated on host" << endl <<endl;

    //initialize the memory allocated on host


    for(int i=0; i < N; i++)
    {
        mySourceGraph_host[i] = 0;
        mySolutionGraph_host[i] = -999;
    }
    foundSolution_host = false;

	
	//************ALLOCATE MEMORY ON CUDA CARD********
	int* mySourceGraph_device; 
	cudaMalloc(&mySourceGraph_device,size); 

	int* mySolutionGraph_device; 
	cudaMalloc(&mySolutionGraph_device,size);

	bool* foundSolution_device; 
	cudaMalloc(&foundSolution_device,sizeof(bool));

    int* answer_device;
    cudaMalloc(&answer_device, (16777216 * sizeof(int)));
    cudaMemset(answer_device, 0, sizeof(*answer_device));
	//************************************************
	
	cout<< "memory allocated on device " <<endl << endl;

 
	//*************TRANSFER FROM HOST TO DEVICE********
	cudaMemcpy(mySourceGraph_device, mySourceGraph_host, size, cudaMemcpyHostToDevice);

	cudaMemcpy(mySolutionGraph_device, mySolutionGraph_host, size, cudaMemcpyHostToDevice);

    cudaMemcpy(foundSolution_device, &foundSolution_host, sizeof(bool), cudaMemcpyHostToDevice);
	//**************************************************

	cout << " transfered memory from host to device " << endl << endl;


    //Determine if this is a CUDA 1.x or 2.x Compute Capability card

    //***********MAJOR.MINOR VERSION NUMBER ***************
	int dev, deviceCount = 0;
    int majorNum;
	cudaError_t error_id = cudaGetDeviceCount(&deviceCount);
	if (error_id != cudaSuccess) 
	{
		printf( "cudaGetDeviceCount returned %d\n-> %s\n", (int)error_id, cudaGetErrorString(error_id) );
	} 
	for (dev = 0; dev < deviceCount; ++dev) 
	{
		cudaDeviceProp deviceProp;
		cudaGetDeviceProperties(&deviceProp, dev);
		cout<<endl;
		printf("  CUDA Capability Major/Minor version number:   %d.%d\n", deviceProp.major, deviceProp.minor);
		majorNum = deviceProp.major;
	}//end for loop
	//*****************************************************

    //*******KERNEL SET-UP***********************
    int numThreads = 0; // The number of threads being used by the GPU.
    //sets the grid,block,thread values depeding on the major and minor numbers
    dim3 grid, block; 
    if (majorNum == 2)//if card is 2.x compute capability, use these parameters for kernel
    {
        grid.x = 32768;
	    grid.y = 1;
	    block.x = 1024;
	    block.y = 1;
        numThreads = 33554432;
    }
    else //otherwise if card is 1.x compute capability, use these parameters
    {
        grid.x = 32768;
	    grid.y = 1;
	    block.x = 512;
	    block.y = 1;
        numThreads = 16777216;
    }
	//***********************************************

	//*******KERNEL CALL ****************************
    Ramsey_Kernel<<<grid, block>>>(mySourceGraph_device, mySolutionGraph_device, foundSolution_device, N, numThreads, answer_device);
	cout<< " kernel called here " << endl << endl;
	//***********************************************


	//*******TRANSFER DATA BACK TO HOST*************
	cudaMemcpy(&foundSolution_host, foundSolution_device, size, cudaMemcpyDeviceToHost);
	//**********************************************
	
	// Check if a graph with no monochromatic K5 was found and if so
	// print it out.
    if (foundSolution_host == true) 
    {
        cout << "Graph found! " << endl;
        cout << " The graph we found is : " << mySolutionGraph_host << endl;
        // Copy the (64 MB - gack!!!) array holding solution from the GPU to the Host
        // This is only ever done if we actually find the graph that proves R(5,5) >= 46
        // so we never expect it to run and therefore won't get a performance hit...
	    cudaMemcpy(answer_host, answer_device, (16777216 * sizeof(int)), cudaMemcpyDeviceToHost);
        
        bool foundAnswerInArray = false;
        int counter = 0;
        while ((foundAnswerInArray == false) && (counter < 16777216))
        {
            if (answer_host[counter] > 0) // We found our answer (aka counter-example showing R(5,5) >= 46
            {
                foundAnswerInArray = true;

		        // Calculate the base graph for this thread.

		        int offset[NUM_EDGES]; // Holds the binary representation of the offset. The offset is used to calculate the 
							           // first graph in each set of graphs.
		        int baseGraph[NUM_EDGES]; // The first graph in the set of graphs to test.
		        int graph[NUM_EDGES]; // The graph in each set of graphs to test.
		        long long remainder; // The offset from the base graph of the first graph in each range.  Used to calcuate 
							         //the binary value of the offset from the base graph.

		        int id = counter; // Set id to the first index of the array that holds an answer.
		        int whichGraph = answer_host[counter] - 1; // which graph was it in the thread that found our graph with no K5.

		        long long graphsPerCore = NUM_GRAPHS/numThreads;

		        // Calculate the first graph for each core to test.

		        // Initializing the arrays
		        for(int j=0; j < NUM_EDGES; j++)
		        {
			        baseGraph[j] = 0; // TODO: Make another version of baseGraph to test a different section of graphs
			        offset[j] = 0;
			        graph[j]=0;
		        }//end for

		        remainder = graphsPerCore * id;

		        // Calculate first graph for each thread to test.

		        // Translates offset into binary and stores the bits into the offset array
		        for(int x = 39; x >= 0; x--)
		        {
			        long long powResult = pow(2.0f, x);
			        if(remainder >= powResult)
			        {
				        offset[989-x]=1;
				        remainder = remainder - powResult;
			        }
			        else 
			        {
				        offset[989-x]=0;
			        }//end if else
		        }//end for

		        // Puts sum of offset and base graph into graph array so that
		        // graph array holds the bits for the first graph to test.
		        for(int m = NUM_EDGES-1; m >= 0; m--)
		        {
			        graph[m]= graph[m]+offset[m]+baseGraph[m];

			        if(graph[m] == 2)
			        {
				        graph[m]=0;
				        graph[m-1] = graph[m-1] +1;
			        }
			        else if ( graph[m]==3)
			        {
				        graph[m]= 1;
				        graph[m-1] = graph[m-1] +1;
        				
			        }//end if else
		        }//end for


		        // reinitialize offset array
		        for(int j=0; j < NUM_EDGES; j++)
		        {
			        offset[j] = 0;

		        }//end for


		        // convert whichGraph to binary and shove it in offset[] array
		        // Translates offset into binary and stores the bits into the offset array
		        remainder = whichGraph;
		        for(int x = 39; x >= 0; x--)
		        {
			        long long powResult = pow(2.0f, x);
			        if(remainder >= powResult)
			        {
				        offset[989-x]=1;
				        remainder = remainder - powResult;
			        }
			        else 
			        {
				        offset[989-x]=0;
			        }//end if else
		        }//end for



		        // Puts sum of offset and base graph into graph array so that
		        // graph array holds the bits for the first graph to test.
		        for(int m = NUM_EDGES-1; m >= 0; m--)
		        {
			        graph[m]= graph[m]+offset[m];

			        if(graph[m] == 2)
			        {
				        graph[m]=0;
				        graph[m-1] = graph[m-1] +1;
			        }
			        else if ( graph[m]==3)
			        {
				        graph[m]= 1;
				        graph[m-1] = graph[m-1] +1;
        				
			        }//end if else
		        }//end for

	            //prints out binary graph
	            for(int i = 0; i < NUM_EDGES; i++)
	            {
		            cout << graph[i];

	            }//end for            
            }
            else
            {
                counter ++;
            }
        }
        
    }
    else
    {
        cout << " No Graph was found. " << endl;

    }//end if else


	//******FREE MEMORY ON HOST**********
	free(mySourceGraph_host);
	free(mySolutionGraph_host);
	free(&foundSolution_host);
    free(answer_host);
	//***********************************

    //******FREE MEMORY ON DEVICE********
	cudaFree(mySourceGraph_device);
	cudaFree(mySolutionGraph_device);
	cudaFree(foundSolution_device);
	cudaFree(answer_device);
	//***********************************


	cout << " memory from host and device freed " << endl << endl;

	//*********END TIMING IN SECONDS***************************
	time_t stop = time(NULL);
	int delta = stop - start;
	cout << "Time taken " << delta <<" seconds " << endl;
	//**********************************************
		
}
//end main




