#ifndef _TOTHRAMSEYTEST_KERNEL_H_
#define _TOTHRAMSEYTEST_KERNEL_H_

#include <stdio.h>

const long long MAX_NUM = 100000000;

const long long NUM_GRAPHS = 335544320000LL; // A long long has a maximum value of 9223372036854775807.
const int RED = 0;
const int BLUE = 1;
const int NUM_VERTICES = 45;
const int NUM_EDGES = 990; //change to 990 for 45 vertices


__global__ void Ramsey_Kernel( int *mySourceGraph_device, int *mySolutionGraph_device, bool *foundSolution_device, int N, int numThreads, int *answer_device) 
{
	// Calculate the base graph for this thread.

    //int numThreads; // ASSUME SET BY PARAMETER PASSED TO KERNEL
	
	int offset[NUM_EDGES]; // Holds the binary representation of the offset. The offset is used to calculate the 
                           // first graph in each set of graphs.
	int baseGraph[NUM_EDGES]; // The first graph in the set of graphs to test.
	int graph[NUM_EDGES]; // The graph in each set of graphs to test.
	long long remainder; // The offset from the base graph of the first graph in each range.  Used to calcuate 
                         //the binary value of the offset from the base graph.
	int id = blockIdx.x * blockDim.x + threadIdx.x;  // IDX FROM CUDA...

	long long graphsPerCore = NUM_GRAPHS/numThreads;

	// Calculate the first graph for each core to test.
	
	// Initializing the arrays
	for(int j=0; j < NUM_EDGES; j++)
	{
		baseGraph[j] = 0; // TODO: Make another version of baseGraph to test a different section of graphs
		offset[j] = 0;
		graph[j]=0;
	}//end for

	remainder = graphsPerCore * id;

	// Calculate first graph for each thread to test.
	
	// Translates offset into binary and stores the bits into the offset array
	for(int x = 39; x >= 0; x--)
	{
		long long powResult = pow(2.0f, x);
		if(remainder >= powResult)
		{
			offset[989-x]=1;
			remainder = remainder - powResult;
		}
		else 
		{
			offset[989-x]=0;
		}//end if else
	}//end for


	// Puts sum of offset and base graph into graph array so that
	// graph array holds the bits for the first graph to test.
	for(int m = NUM_EDGES-1; m >= 0; m--)
	{
		graph[m]= graph[m]+offset[m]+baseGraph[m];

		if(graph[m] == 2)
		{
			graph[m]=0;
			graph[m-1] = graph[m-1] +1;
		}
		else if ( graph[m]==3)
		{
			graph[m]= 1;
			graph[m-1] = graph[m-1] +1;
			
		}//end if else
	}//end for

	// Test all the graphs in the range.

	// Tests the first graph.

	//bool result = hasK5(graph);
    //////////////////////////////////////////////////////////////////////////////////////////
    // We replaced the hasK5() function call with the code.
    bool result;

    // Begin hasK5() inlining.
    bool foundK5 = false;

    // initialize counters
    int vertex1 = 0;
    int vertex2 = 1;
    int vertex3 = 2;
    int vertex4 = 3;
    int vertex5 = 4;

    /* These five nested while loops select the 5 vertices that 
       we will test to see if they form a K5. The outermost while loop
       selects the smallest numbered vertex.The next while loop selets the second smallest
       numbered vertex and so on... The set of while loops test every possible
       set of five vertices that could form a K5 until it finds one or has tested every possible
       set.
    */

    while (!foundK5 && vertex1 < NUM_VERTICES - 4)
    {
        while (!foundK5 && vertex2 < NUM_VERTICES - 3)
        {
	        while (!foundK5 && vertex3 < NUM_VERTICES - 2)
	        {
		        while (!foundK5 && vertex4 < NUM_VERTICES - 1)
		        {
			        while (!foundK5 && vertex5 < NUM_VERTICES)
			        {
				        if (graph[((NUM_VERTICES * vertex1) + vertex2 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
					        graph[((NUM_VERTICES * vertex1) + vertex3 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
					        graph[((NUM_VERTICES * vertex1) + vertex4 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
					        graph[((NUM_VERTICES * vertex1) + vertex5 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&							
					        graph[((NUM_VERTICES * vertex2) + vertex3 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
					        graph[((NUM_VERTICES * vertex2) + vertex4 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
					        graph[((NUM_VERTICES * vertex2) + vertex5 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
					        graph[((NUM_VERTICES * vertex3) + vertex4 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == RED &&							
					        graph[((NUM_VERTICES * vertex3) + vertex5 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == RED &&
					        graph[((NUM_VERTICES * vertex4) + vertex5 - (((vertex4 + 1)*(vertex4 + 2))/ 2))] == RED)
				        {
					        foundK5 = true;
				        }
				        else if (graph[((NUM_VERTICES * vertex1) + vertex2 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE&&
					        graph[((NUM_VERTICES * vertex1) + vertex3 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex1) + vertex4 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex1) + vertex5 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex2) + vertex3 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex2) + vertex4 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex2) + vertex5 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex3) + vertex4 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex3) + vertex5 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == BLUE &&
					        graph[((NUM_VERTICES * vertex4) + vertex5 - (((vertex4 + 1)*(vertex4 + 2))/ 2))] == BLUE)
				        {
					        foundK5 = true;
				        }
				        vertex5++;
			        }
			        vertex4++;
			        vertex5 = vertex4 + 1;
		        }
		        vertex3++;
		        vertex4 = vertex3 + 1;
		        vertex5 = vertex4 + 1;
	        }
	        vertex2++;
	        vertex3 = vertex2 + 1;
	        vertex4 = vertex3 + 1;
	        vertex5 = vertex4 + 1;
        }
        vertex1++;
        vertex2 = vertex1 + 1;
        vertex3 = vertex2 + 1;
        vertex4 = vertex3 + 1;
        vertex5 = vertex4 + 1;
    }

    if (foundK5 == true)
    {
        result = true;
    }
    else
    {
        result = false;
    }
    // End hasK5() inlining.
    //////////////////////////////////////////////////////////////////////////////////////////

	// If a graph with no K5 is found, print out that graph.
	if(result == false)
	{
        *foundSolution_device = true;
        answer_device[id] = 1;
	}//end if

	long long loopIndex = 0LL;

	 // This is a for loop that tests all graphs for this core.
	for(loopIndex = 0; loopIndex < graphsPerCore; loopIndex++)
	{
		// Generate next graph.

        // We have replaced the function call for generateNextGraph() with the code.
		//generateNextGraph(graph);
        //////////////////

        bool carry = false;
        int index = NUM_EDGES -1;	

        do 
        {
            //add 1 and check for carry 
            if ( graph[index] == 0)
            {
	            carry = false;
	            graph[index] = 1; //set to 1
            }
            else
            {
	            carry = true;
	            graph[index] = 0; //set to 0
            }//end if/else

            index = index - 1;
        }
        while((carry == true) && (index >= 0));

        index = NUM_EDGES - 1;
        //////////////////


		// Test graph.
        //result = hasK5(graph);

        //////////////////////////////////////////////////////////////////////////////////////////
        // We replaced the hasK5() function call with the code.

        // Begin hasK5() inlining.
        bool foundK5 = false;

        // initialize counters
        int vertex1 = 0;
        int vertex2 = 1;
        int vertex3 = 2;
        int vertex4 = 3;
        int vertex5 = 4;

        /* These five nested while loops select the 5 vertices that 
           we will test to see if they form a K5. The outermost while loop
           selects the smallest numbered vertex.The next while loop selets the second smallest
           numbered vertex and so on... The set of while loops test every possible
           set of five vertices that could form a K5 until it finds one or has tested every possible
           set.
        */

        while (!foundK5 && vertex1 < NUM_VERTICES - 4)
        {
	        while (!foundK5 && vertex2 < NUM_VERTICES - 3)
	        {
		        while (!foundK5 && vertex3 < NUM_VERTICES - 2)
		        {
			        while (!foundK5 && vertex4 < NUM_VERTICES - 1)
			        {
				        while (!foundK5 && vertex5 < NUM_VERTICES)
				        {
					        if (graph[((NUM_VERTICES * vertex1) + vertex2 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
						        graph[((NUM_VERTICES * vertex1) + vertex3 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
						        graph[((NUM_VERTICES * vertex1) + vertex4 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&
						        graph[((NUM_VERTICES * vertex1) + vertex5 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == RED &&							
						        graph[((NUM_VERTICES * vertex2) + vertex3 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
						        graph[((NUM_VERTICES * vertex2) + vertex4 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
						        graph[((NUM_VERTICES * vertex2) + vertex5 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == RED &&							
						        graph[((NUM_VERTICES * vertex3) + vertex4 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == RED &&							
						        graph[((NUM_VERTICES * vertex3) + vertex5 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == RED &&
						        graph[((NUM_VERTICES * vertex4) + vertex5 - (((vertex4 + 1)*(vertex4 + 2))/ 2))] == RED)
					        {
						        foundK5 = true;
					        }
					        else if (graph[((NUM_VERTICES * vertex1) + vertex2 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE&&
						        graph[((NUM_VERTICES * vertex1) + vertex3 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex1) + vertex4 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex1) + vertex5 - (((vertex1 + 1)*(vertex1 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex2) + vertex3 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex2) + vertex4 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex2) + vertex5 - (((vertex2 + 1)*(vertex2 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex3) + vertex4 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex3) + vertex5 - (((vertex3 + 1)*(vertex3 + 2))/ 2))] == BLUE &&
						        graph[((NUM_VERTICES * vertex4) + vertex5 - (((vertex4 + 1)*(vertex4 + 2))/ 2))] == BLUE)
					        {
						        foundK5 = true;
					        }
					        vertex5++;
				        }
				        vertex4++;
				        vertex5 = vertex4 + 1;
			        }
			        vertex3++;
			        vertex4 = vertex3 + 1;
			        vertex5 = vertex4 + 1;
		        }
		        vertex2++;
		        vertex3 = vertex2 + 1;
		        vertex4 = vertex3 + 1;
		        vertex5 = vertex4 + 1;
	        }
	        vertex1++;
	        vertex2 = vertex1 + 1;
	        vertex3 = vertex2 + 1;
	        vertex4 = vertex3 + 1;
	        vertex5 = vertex4 + 1;
        }

        if (foundK5 == true)
        {
	        result = true;
        }
        else
        {
	        result = false;
        }
        // End hasK5() inlining.
        //////////////////////////////////////////////////////////////////////////////////////////

		// If a graph with no K5 is found, print out that graph.
		if(result == false)
		{
            *foundSolution_device = true;
            answer_device[id] = loopIndex+1;
		}//end if
	}//end for loop

}

#endif


